# -*- coding: japanese-shift-jis-dos; tab-width: 4; -*-
#
# wdic-body - FreePWING script for WDIC 2002
# $Id: wdic-body.pl 1.3.1.0 2002/08/14 08:41:54 satomii Exp $
#
# This file is designed to be used from the wdic-fpw.pl script.
# DO NOT CHANGE THE KANJI CODE OF THIS FILE.
#
# Copyright (C) 2000-2002, rei@wdic.org.
# This program is distributed in the hope that it will be useful, but
# without any warranty. See the GNU General Public License for the details.
#
require 5.005;

use strict;
use IO::File;
use File::DosGlob 'glob';
use Jcode;

# use Unicode::Japanese for handling UTF-16 file, specifically
# yoshigyu.txt. if --opt-plugin is not set, this is not used. reading from
# UTF-16 file is NOT TESTED at all, since Unicode::Japanese could not be
# installed properly on Win32/VC++ without modification. it is recommended
# to manually convert the file(s) before running wdic-fpw.
eval { require Unicode::Japanese; };

my $PLUGIN_HEADING_SUFFIX = Jcode::convert('yvOCz', 'euc');
my $PLUGIN_IMAGE_LABEL = Jcode::convert('[}]', 'euc');

#
# void make_body($word_count)
# the entry point to generate word entries.
#
sub make_body($)
{
	my $word_count = $_[0];
	my ($i, $key);

	print "Writing word entries...\n";

	for ($i = 0; $i < $word_count; $i++) {
		my %word = $wdic->define_word($i);
		&add_word(\%word);
	}

	$i += &write_plugin_files() if ($opt_plugin);
	$i += &write_plugin_images() if ($opt_cgraphs);
	$i += &write_plugin_ext();
	print "Total $word_count words, $i entries written.\n";
}

#
# (int, int) new_entry(char *tag, int offset)
# switch to the new entry and writes a tag. it is recommended to add
# FreePWING::BaseText::add_entry_tag method to your FreePWING package
# by applying the patch from [edict 1364].
#
sub new_entry($;$)
{
	my($tag, $offset) = @_;
	my($text_pos, $head_pos) = (0, 0);

	$fpwtext->new_entry()
		|| die &death_cry("Unable to start new entry", $fpwtext);
	$fpwheading->new_entry()
		|| die &death_cry("Unable to start new entry", $fpwheading);
	$text_pos = $fpwtext->entry_position();
	$head_pos = $fpwheading->entry_position();

	$tag .= sprintf("%08x", $offset) if (defined($offset));

	if ($fpwtext->can('add_entry_tag')) {
		$fpwtext->add_entry_tag($tag)
			|| die &death_cry("Unable to add entry tag: $tag", $fpwtext);
	} else {
		$fpwtext->add_tag($tag)
			|| die &death_cry("Unable to add entry tag: '$tag", $fpwtext);
	}
	return ($head_pos, $text_pos, $tag);
}

##-------------------------------------------------------------------------
## WORD ENTRY

#
# void add_word((hash) word)
# adds one word entry to the FreePWING context.
#
sub add_word(%)
{
	my %word = %{ $_[0] };
	my ($entry_tag, $text_pos, $head_pos);
	my (@list, $word, $group_name, $file_name, $index);

	printf("$word{GROUP}:$word{FILE}(%08x): %s",
		$word{OFFSET}, Jcode::convert($word{WORD}, $opt_charset, 'sjis'));
	$newline_on_death = 1;

	$word = $word{WORD};
	if ($word =~ s/(\s+\([0-9]+\))$//) { $index = $1; }
	else { $index = ''; }

	$group_name = $wdic->get_group_title($word{GROUP});
	$group_name =~ s/̊bm$//;
	$file_name = &reform_sjis($wdic->get_file_title($word{FILE}), 0);
	$file_name =~ s/ //g;
	$file_name = ":$file_name" if ($group_name);

	# switch to the new entry and write the heading.
	($head_pos, $text_pos, $entry_tag) = &new_entry($word{FILE}, $word{OFFSET});
	$_ = $group_name;
	$_ =~ s/p$//;
	$_ .= $file_name;
	$_ =~ s/$//;
	&write_sjis($fpwheading, "$word{WORD} y$_z");

	# register search words.
	$_ = &reform_sjis($word, 2);
	if (!/^[\s\-\']*$/) {
		$_ = Jcode::convert($_, 'euc', 'sjis');
		$fpwword2->add_entry($_, $head_pos, $text_pos)
			|| die &death_cry("Unable to add search word: $_", $fpwword2);
	}
	push(@list, $word, $word{FILE});

	if ($word{READ} !~ /^\?+$/) {
		$fpwword2->add_entry(Jcode::convert($word{READ}, 'euc', 'sjis'),
				$head_pos, $text_pos)
			|| die &death_cry("Unable to add search word: $word{READ}",
				$fpwword2);
		push(@list, &make_reading_keywords($word{READ}));
	}

	push(@list, $word{GROUP}) if ($word{GROUP} =~ /^[A-Z]+$/i);
	push(@list, &make_english_keywords($head_pos, $text_pos, $word{ENGLISH}));
	push(@list, &make_link_keywords(@{ $word{REFERENCE} }));
	push(@list, @{ $word{DIRS} }) if ($word{DIRS});
	&add_keyword_list($head_pos, $text_pos, @list);
	undef @list;

	# write the first line.
	if (!$fpwtext->add_keyword_start()
			|| !$fpwtext->add_text(Jcode::convert($word, 'euc', 'sjis'))
			|| !$fpwtext->add_keyword_end()) {
		die &death_cry("Unable to write the first line", $fpwtext);
	}
	&write_sjis($fpwtext, "$index y$group_name$file_namez\n");

	# write the next line which may contain word reading, english,
	# pronounce and POS.
	&write_sjis($fpwtext, "[$word{READ}]");
	&write_sjis($fpwtext, " ($word{ENGLISH})") if ('' ne $word{ENGLISH});
	&write_phonetic($word{PRONOUNCE}) if ($word{PRONOUNCE});
	&write_sjis($fpwtext, " k$word{POS}l") if ('' ne $word{POS});
	&write_newline($fpwtext, 1);

	# write the directory (category) lines. still TBD what should be done
	# here. should make a list of directories later?
	if ($opt_dirinfo && $word{DIRS}) {
		my $name;
		foreach (@{ $word{DIRS} }) {
			$fpwtext->add_text($_)
				|| die &death_cry("Unable to add dir: $_", $fpwtext);
			$name = $wdic->get_dir_name($_);
			if ($name ne $_) {
				$name = &reform_sjis($name, 0);
				&write_sjis($fpwtext, " c $name");
			}
			&write_newline($fpwtext, 1);
		}
	}

	# write definition lines.
	@list = @{ $word{DEFINITION} };
	if (@list) {
		foreach (@list) {
			warn "Unexpected format: $_" if (!s/^E//);
			$fpwtext->add_text($opt_bullet)
				|| die &death_cry("Failed to add text: $opt_bullet", $fpwtext);
			$_ = &reform_sjis($_, 0);
			s/\" , \"/\", \"/;
			s/($SJIS_ANY)(\.[zZ] )/$1 $2/g;
			s/(gq)(\.[0-9a-zA-Z\.]+)/$1 $2/g;
			s/(\<p\>) (.+) (\<\/p\>)/$1$2$3/;
			s/C/b/g;
#			s/(\x82[\x9f-\xf1])([\-\+][0-9a-zA-Z,\.]+)([ \)]$SJIS_ANY)/$1 $2$3/;
#			s/(\x83[\x40-\x96])([\-\+][0-9a-zA-Z,\.]+)([ \)]$SJIS_ANY)/$1 $2$3/;
			s/([\x88-\xfc][\x40-\x7e\x80-\xfc])([\-\+][0-9a-zA-Z,\.]+)([ \)]$SJIS_ANY)/$1 $2$3/;
			s/\(pc  /(pc/;
			s/($SJIS_ANY)(\?\?\?+\.[a-z]+)/$1 $2/;	# '???.to' domain.
			s/()(\*)(y)/$1 $2 $3/;				# in 'DTMF'.
			s/(Richard M\.)(Stallman)/$1 $2/g;
			&write_sjis($fpwtext, "$_\n");
		}
	}
	$fpwtext->add_indent_level(2);

	# write columns and/or examples, if any.
	@list = @{ $word{COLUMN} };
	if (@list) {
		my $image_block = 0;
		&write_newline($fpwtext, 1);
		foreach (@list) {
			if (/^/) {
				$image_block = 1;
			} elsif (/^$/) {
				$image_block = 0;
			} elsif ($image_block && /^/) {
				&write_link_list('', '', $_);
				next;
			}
			&write_sjis($fpwtext, "$_\n");
		}
	}

	# write links.
	@list = @{ $word{REFERENCE} };
	if (@list) {
		&write_newline($fpwtext, 1) if (@{ $word{DEFINITION} });
		&write_link_list($word{WORD}, $word{GROUP}, @list);
	}

	# and add a newline at the end of this entry.
	&write_newline($fpwtext, 1);
	$newline_on_death = 0;
	print "\n";
}

##-------------------------------------------------------------------------
## KEYWORDS

#
# void add_keyword_list(char **euc_keywords)
# registers each keywords in the given list to the current word entry.
#
sub add_keyword_list($$@)
{
	my($head_pos, $text_pos, @key) = @_;
	my($key, $prev_key);

	foreach $key (@key) {
		# do not use DBCS alphabets - due to FreePWING behavior (bug?).
		$key = &reform_sjis($key, 2);
		$key = Jcode::convert($key, 'euc', 'sjis');
		$key = uc($key);
	}

	@key = sort(@key);
	$prev_key = '';

	foreach $key (@key) {
		next if ($key =~ /^[\s\-\']*$/ || $key eq $prev_key);
		$fpwkeyword->add_entry($key, $head_pos, $text_pos)
			|| die &death_cry("Unable to add keyword: $key", $fpwkeyword);
		$prev_key = $key;
	}
}

#
# char **make_reading_keywords(char *read)
#
sub make_reading_keywords($)
{
	my $read = $_[0];
	my(@key, $key, $c);

	$key = '';
	while ('' ne $read) {
		$c = substr($read, 0, 1);
		$read = substr($read, 1);
		if ($c =~ /[\x81-\x9f\xe0-\xfc]/) {
			$c .= substr($read, 0, 1);
			$read = substr($read, 1);
		}
		if ('@' eq $c || 1 == length($c)) {	# '@' or '-'.
			push(@key, $key) if ('' ne $key);
			$key = '';
		} else {
			$key .= $c;
		}
	}
	push(@key, $key) if ('' ne $key);
	return @key;
}

#
# char **make_english_keywords(int heading_pos, int text_pos, char *english)
# registers english terms as search words. also returns the keyword list
# for 'conditional' search.
#
sub make_english_keywords($$$)
{
	my($head_pos, $text_pos, $english) = @_;
	my(%keyword, @key);

	$english = &reform_sjis($english, 2);
	$english = Jcode::convert($english, 'euc', 'sjis');
	s/ *\[[^\]]+\]//g;
	s/\\/\xa1\xef/g;		# '' in EUC (a1ef).

	# search words.
	@key = split(/[:;\/] +/, $english);
	foreach (@key) {
		s/^\s+//;
		s/\s+$//;
		if (!/^[\-\']*$/ && !$keyword{$_}) {
			$fpwword2->add_entry($_, $head_pos, $text_pos)
				|| die &death_cry(
					"Cannot add search word: $_ (in \"$english\")", $fpwword2);
			$_ = uc($_);
			$keyword{$_} = 1;
		}
	}

	# jouken-kensaku keywords.
	$_ = Jcode::convert(join(' ', @key), 'sjis', 'euc');
	return split(/\s+/);
}

#
# char **make_link_keywords(char **link_list)
#
sub make_link_keywords(@)
{
	my(@key, $key, $str, $c);

	foreach (@_) {
		next if (!/^\x81[\xa8\x9f\x9a]/);	# ![], not a 'word'.
		$str = $_;
		$str =~ s/^\x81.//;
		$str =~ s/ +:[A-Z]+$//i;
		$str =~ s/ +\([0-9]+\)$//;
		push(@key, $str);

		$key = '';
		while ('' ne $str) {
			$c = substr($str, 0, 1);
			$str = substr($str, 1);
			if ($c =~ /[\x81-\x9f\xe0-\xfc]/) {
				$c .= substr($str, 0, 1);
				$str = substr($str, 1);
			}
			if ('@' eq $c) {
				push(@key, $key) if ('' ne $key);
				$key = '';
			} else {
				$key .= $c;
			}
		}
		push(@key, $key) if ('' ne $key && $_ ne $key);
	}

	return @key;
}

##-------------------------------------------------------------------------
## REFERENCES

#
# void write_link_list(char *current_word, char *current_group, char **list)
# writes a list of related words, making cross-reference available if
# the target word is found in the loaded words.
#
sub write_link_list($$@)
{
	my $current_word = shift(@_);
	my $current_group = shift(@_);
	my ($bullet, $file, $ext, $key, $target_group, $group);

	foreach (@_) {
		if (s/^([^ \.\/\\<>\|\*\?]+)\.([a-z]+) *//i) {
			# plugin file.
			($file, $ext, $key) = ($1, $2, "$2_$1");
			$_ = &reform_sjis($_, 0);

			if (defined($cgraphs{$key})) {
				if ($cgraphs{$key}{name}) {
					push(@{ $cgraphs{$key}{list} }, "$current_word");
				} else {
					$cgraphs{$key}{name} = "$file.$ext";
					$cgraphs{$key}{list} = [ "$current_word" ];
				}
				$_ = "$file.$ext" . ($_ ? " $_" : '');
				if (!$fpwtext->add_color_graphic_start($key)
						|| !&write_sjis($fpwtext, $_)
						|| !$fpwtext->add_color_graphic_end()) {
					die &death_cry("Unable to add DIB: $file: $_", $fpwtext);
				}

			} elsif ($opt_plugin && $ext =~ /^(csv|doc|txt)$/i) {
				$key = "$file.$ext";
				$file = glob("$opt_plugin/$key");
				if (!$file && -f "$opt_plugin/$key") {
					$file = "$opt_plugin/$key";
				}
				if ($file) {
					if (defined($plugins{$key})) {
						push(@{ $plugins{$key}{list} }, "$current_word");
					} else {
						$plugins{$key}{tag} = sprintf("PLUGIN%04x",
								scalar(keys(%plugins)) + 1);
						$plugins{$key}{path} = $file;
						$plugins{$key}{list} = [ "$current_word" ];
					}
					&write_sjis($fpwtext, '');
					&write_link($fpwtext, "$key", $plugins{$key}{tag});
					&write_sjis($fpwtext, " $_") if ($_);
				} else {
					$_ = "$key" . ($_ ? " $_" : '');
					&write_sjis($fpwtext, $_);
				}

			} else {
				$_ = "$file.$ext" . ($_ ? " $_" : '');
				&write_sjis($fpwtext, $_);
			}

		} elsif (s/^(\x81[\xa8\x9f\x9a])//) {
			# [], link to another entry.
			$bullet = $1;
			$target_group = (s/ :([A-Z]+)$//i) ? $1 : $current_group;
			($group, $file) = &write_word_link($fpwtext,
					$bullet, $_, $target_group, '');
			$group = $wdic->get_group_title($group ? $group : $target_group);
			$group =~ s/̊bm$//;
			$group =~ s/p$//;
			if ($file) {
				$file = $wdic->get_file_title($file);
				$file =~ s/$//;
				$file = &reform_sjis($file, 0);
				$group .= ':' if (group);
				$group .= $file;
			}
			&write_sjis($fpwtext, " s$groupt") if ($group);

		} else {
			# not [], outside information.
			if (s/^([^ ]+)(.*)$/$1/) {	# URL.
				$_ .= &reform_sjis($2, 0) if ($2);
			} elsif (/^/) {				# URL group heading.
				$_ = &reform_sjis($_, 0);
			}
			&write_sjis($fpwtext, $_);
		}

		&write_newline($fpwtext, 1);
	}
}

##-------------------------------------------------------------------------
## PHONETIC SYMBOLS

#
# void write_phonetic(char *pron);
# writes a pronounce into the current context using or not using user
# characters.
#
sub write_phonetic($)
{
	my $pron = " /$_[0]/";
	my($c, @gaiji);

	$pron = Jcode::convert($pron, 'euc', 'sjis') if ($pron =~ /[^0x20-0x7e]/);

	if (!$opt_gaiji) {
		# do not use gaiji - just write as normal text.
		$fpwtext->add_text($pron)
			|| die &death_cry("Unable to add pronounce: $_[0]", $fpwtext);
		return;
	}

	while ('' ne $pron) {
		$c = substr($pron, 0, 1);
		$pron = substr($pron, 1);
		undef @gaiji;
		if ($c =~ /[\/\|eiou]/) {
			# nop.
		} elsif ($c =~ /[3S]/) {
			$gaiji[0] = "phon-${c}";
		} elsif ('0' eq $c) {
			$gaiji[0] = 'phon-0' if ($pron =~ s/^\-//);
		} elsif ('a' eq $c) {
			@gaiji = ('phon-ae1', 'phon-ae2') if ($pron =~ s/^e//);
		} elsif ('n' eq $c) {
			$gaiji[0] = 'phon-n' if ($pron =~ s/^,//);
		} elsif ('(' eq $c) {
			$gaiji[0] = 'phon-r' if ($pron =~ s/^r\)//);
		} elsif ('\'' eq $c) {
			if ($pron =~ s/^ae//) {
				@gaiji = ('phon-ae3', 'phon-ae4');
			} elsif ($pron =~ s/^([aeioSu])//) {
				$gaiji[0] = "phon-${1}u";
			} elsif ($pron =~ s/^\#//) {
				if ($pron =~ s/^([cev])//) {
					$gaiji[0] = "phon-r${1}u";
				} else {
					$c = '?';
				}
			} elsif ($pron =~ s/^6//) {
				$gaiji[0] = 'phon-6';
			} else {
				# &#768;
			}
		} elsif ('`' eq $c) {
			if ($pron =~ s/^ae//) {
				@gaiji = ('phon-ae5', 'phon-ae6');
			} elsif ($pron =~ s/^([aeioSu])//) {
				$gaiji[0] = "phon-${1}d";
			} elsif ($pron =~ s/^\#//) {
				if ($pron =~ s/^([cev])//) {
					$gaiji[0] = "phon-r${1}d";
				} else {
					$c = '?';
				}
			} else {
				# &#769;
			}
		} elsif ('#' eq $c) {
			if ($pron =~ s/^e\+//) {
				$gaiji[0] = 'phon-rei';
			} elsif ($pron =~ s/^([ev])//) {
				$gaiji[0] = "phon-r${1}";
			} else {
				$c = '?';
			}
		} elsif ('[' eq $c) {
			if ($pron =~ s/^([^\]]+\])//) { $c .= $1; }
		}
		if (@gaiji) {
			$fpwtext->add_half_user_character($gaiji[0])
				|| die &death_cry(
					"Unable to add pronounce: $_[0] ($gaiji[0])", $fpwtext);
			if ($gaiji[1]) {
				$fpwtext->add_half_user_character($gaiji[1])
					|| die &death_cry(
						"Unable to add pronounce: $_[1] ($gaiji[1])", $fpwtext);
			}
		} else {
			$fpwtext->add_text($c)
				|| die &death_cry(
					"Unable to add pronounce: $_[0] ($c)", $fpwtext);
		}
	}
}

##-------------------------------------------------------------------------
## PLUGIN FILE ENTRY

#
# void write_plugin_files(void)
# creates text entries for text-based plugin files if necessary.
#
sub write_plugin_files()
{
	my @keys = keys(%plugins);
	my ($key, $tag, $file, $head_pos, $text_pos, $code);

	return if (!scalar(@keys));
	@keys = sort @keys;
	print "Writing plugin entries...\n";

	foreach $key (@keys) {
		$tag = $plugins{$key}{tag};
		open(IN, "<$plugins{$key}{path}")
			|| die "Unable to open file: $plugins{$key}{path} ($!)";

		printf("$tag: $key\n");
		$fpwtext->new_context()
			|| die &death_cry("Unable to start new context", $fpwtext);
		($head_pos, $text_pos) = &new_entry($tag);

		$fpwheading->add_text("$key $PLUGIN_HEADING_SUFFIX")
			|| die &death_cry("Unable to add heading: $key", $fpwheading);
		$fpwword2->add_entry($key, $head_pos, $text_pos)
			|| die &death_cry("Unable to add search word: $key", $fpwword2);

		if (!$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($key)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_newline()) {
			die &death_cry("Unable to write the first line", $fpwtext);
		}

		&write_link_list('', '', @{ $plugins{$key}{list} });
		&write_newline($fpwtext, 1);
		$fpwtext->add_indent_level(2);

		while (<IN>) {
			s/[\r\n\x1a]+$//;
			if ($_) {
				# plugin may not be written in SJIS. always convert it to
				# sjis so that single-byte kana's are handled properly.
				$code = Jcode::getcode($_);
				if ('ucs2' eq $code || 'binary' eq $code) {
					# might be utf16, which is not supported by Jcode.
					# THIS CODE IS NOT TESTED AT ALL.
					eval {
						my $u = Unicode::Japanese->new($_);
						$_ = $u->sjis();
					};
					s/[^\x20-\x7e]/./g if ($@);
				} elsif ('ascii' ne $code && 'sjis' ne $code) {
					$_ = Jcode::convert($_, 'sjis');
				}
				&write_sjis($fpwtext, $_);
			}
			&write_newline($fpwtext, 1);
		}

		close(IN);
	}

	return scalar(@keys);
}

#
# void write_plugin_images(void)
# creates text entries for plugin images.
#
sub write_plugin_images()
{
	my @keys = keys(%cgraphs);
	my ($key, $tag, $file, $head_pos, $text_pos);

	return if (!scalar(@keys));
	@keys = sort @keys;
	print "Writing graphic entries...\n";

	$fpwtext->new_context()
		|| die &death_cry("Unable to start new context", $fpwtext);

	foreach $key (@keys) {
		next if (!$cgraphs{$key}{name});
		$tag = "PLUGIN_$key";
		$file = $cgraphs{$key}{name};

		printf("$tag: $file\n");
		($head_pos, $text_pos) = &new_entry($tag);

		$fpwheading->add_text("$file $PLUGIN_HEADING_SUFFIX")
			|| die &death_cry("Unable to add heading: $file", $fpwheading);
		$fpwword2->add_entry($file, $head_pos, $text_pos)
			|| die &death_cry("Unable to add search word: $file", $fpwword2);

		if (!$fpwtext->add_keyword_start()
				|| !$fpwtext->add_text($file)
				|| !$fpwtext->add_keyword_end()
				|| !$fpwtext->add_newline()) {
			 die &death_cry("Unable to write the first line", $fpwtext);
		}
		if (!$fpwtext->add_color_graphic_start($key)
				|| !$fpwtext->add_text($PLUGIN_IMAGE_LABEL)
				|| !$fpwtext->add_color_graphic_end()
				|| !$fpwtext->add_newline()) {
			die &death_cry("Unable to add graphic: $key ($file)", $fpwtext);
		}
		if ($cgraphs{$key}{list}) {
			&write_link_list('', '', @{ $cgraphs{$key}{list} });
			&write_newline($fpwtext, 1);
		}
	}

	return scalar(@keys);
}

#
# int write_plugin_ext(void);
# appends 'ext.csv' plug-in if it exists in the source directory.
#
sub write_plugin_ext()
{
	my %ftype = {
		BINARY => 'oCi',
		TEXT => 'eLXg'
	};
	my %fcnt = {
		AUDIO => 'yEȂ',
		IMAGE => '摜',
		VIDEO => '',
		COMP => 'kf[^',
		ARC => 'k𔺂ȂA[JC',
		PROG => 'vO\[Xt@C',
		EXE => 'vOst@C',
		INF => '',
		DB => 'f[^x[X',
		TEXT => ''
	};
	my $ext = ' ';
	my (@col, $file, $handle, $text_pos, $head_pos, $cnt, $tag);

	return 0 if (!$opt_plugin);

	$file = File::DosGlob::glob("$opt_plugin/ext.csv");
	$file = CORE::glob("$opt_plugin/ext.csv") if (!$file);
	$handle = new IO::File($file, 'r') if ($file);
	return 0 if (!defined($handle));
	binmode $handle;
	print "Plugin file 'ext.csv' found, start parsing...\n";

	$fpwtext->new_context()
		|| die &death_cry("Unable to start new context", $fpwtext);

	while ($_ = $handle->getline()) {
		s/([^\r\n]*)[\r\n]+$/$1/;
		@col = split(/,/);
		next if ('' eq $col[0]);

		$cnt++;
		printf("EXT(%08x): $col[0]\n", $cnt);

		# add this line as a new entry.
		($head_pos, $text_pos) = &new_entry('EXT', $cnt);
		&write_sjis($fpwheading, "$col[0] ygqz");

		# register search words.
		if (!$fpwword2->add_entry($col[0], $head_pos, $text_pos)
				|| !$fpwword2->add_entry(".$col[0]", $head_pos, $text_pos)) {
			die &death_cry("Unable to add search word: $col[0]", $fpwword2);
		}

		# write the first line.
		$fpwtext->add_keyword_start();
		&write_sjis($fpwtext, $col[0]);
		$fpwtext->add_keyword_end();
		write_sjis($fpwtext, " ygqz\n");
		$fpwtext->add_indent_level(2);

		# file type.
		if ('' ne $col[1]) {
			if (defined($ftype{$col[1]})) { $_ = $ftype{$col[1]}; }
			else { $_ = $col[1]; }
		} else {
			$_ = '';
		}
		if ('' ne $col[2]) {
			$_ .= '/' if ('' ne $_);
			if (defined ($fcnt{$col[2]})) { $_ .= $fcnt{$col[2]}; }
			else { $_ .= $col[2]; }
		}
		if ('' ne $_) {
			$fpwtext->add_text($opt_bullet)
				|| die &death_cry("Unable to add text: $opt_bullet", $fpwtext);
			&write_sjis($fpwtext, "$_\n");
		}

		# MIME type.
		if ('' ne $col[3]) { $_ = $col[3] }
		else { $_ = ''; }
		if ('' ne $col[4]) {
			$_ .= '/' if ('' ne $_);
			$_ .= $col[4];
		}
		if ('' ne $_) {
			$fpwtext->add_text($opt_bullet)
				|| die &death_cry("Unable to add text: $opt_bullet", $fpwtext);
			&write_sjis($fpwtext, "MIME ^Cv: $_\n");
		}

		# description.
		if ('' ne $col[5]) {
			$fpwtext->add_text($opt_bullet)
				|| die &death_cry("Unable to add text: $opt_bullet", $fpwtext);
			&write_sjis($fpwtext, &reform_sjis($col[5], 0) . "\n");
		}

		&write_newline($fpwtext, 1);
	}

	$handle->close();
	return $cnt;
}

1;
